const 
  BYTE = 1,
  CHAR = 1,
  USHORT = 2,
  SHORT = 2,
  FWORD = 2,
  UFWORD = 2,
  F2DOT14 = 2,
  UINT24 = 3,
  ULONG = 4,
  LONG = 4,
  LONGDATETIME = 8,
  FIXED = 4;


var sizeIds = {};
var fontData;

var mbLookup = {};

/*
structure.prototype.collapse = function(data, length)
{
  var new_new_data, new_data, 
    capacity = 1024, length, sz;
    
  for(var i = 0; sz = this.sizes[i]; i++)
  {
    for(var j = 0; j < sz.copies; j++)
    {
      while(length + sz.length > capacity)
      {
        capacity *= 2;
        new_new_data = 
      }
    }
  }
  
  for (i = 0; i < structure->cur_length; i++)
  {
    struct size *sz = &structure->sizes[i];
    uint32_t j;
    for (j = 0; j < sz->copies; j++)
    {
      while (*length + sz->length > capacity)
      {
        capacity *= 2;
        new_new_data = (char*)malloc(capacity);
        memcpy(new_new_data, new_data, *length);
        free(new_data);
        new_data = new_new_data;
        ptr = new_data + (*length);
      }

      memcpy(ptr, data + sz->offset, sz->length);
      ptr += sz->length;
      (*length) += sz->length;
    }
  }
  
  if (structure->collapse_fn != NULL)
    structure->collapse_fn(new_data, *length);
  
  return new_data;


function getLong(data, offset, bigEndian)
{
  return ls(data.charCodeAt(offset), 24) + ls(data.charCodeAt(offset+1), 16) + ls(data.charCodeAt(offset+2), 8) + data.charCodeAt(offset+3);
}

function getShort(data, offset, bigEndian)
{
  return ls(data.charCodeAt(offset), 8) + data.charCodeAt(offset+1);
}
*/

var success = 0;
var total = 0;

function getMBInt(length, data, offset)
{
  var subdata = data.substr(offset, length);
  if(subdata in mbLookup)
    return mbLookup[subdata]; //use a lookup table because it's cheaper to do so
  else
  {
    var ret = 0;
    for(var i = 0; i < length; i++)
      ret += ls(data.charCodeAt(offset + length - 1 - i), i*8);
      
    mbLookup[subdata] = ret;
    return ret;
  }
}

function rand_range(min, max)
{
  //[min, max)
  if(max - min > 0)
    return Math.floor(Math.random() * (max - min)) + min;
  
  return 1;
}

function rand_max(max)
{
  //[0, max)
  return Math.floor(rand_range(0, max));
}


function assert(condition)
{
  if(condition)
  {
    if(window.console)
      console.log("assertion failure");
    //debugger;
  }
}

function ls(i, bits)
{
  //left bitwise shift
  if(i > 0xfffffff)
    return i * Math.pow(2, bits);
  else
    return i << bits;
}

function rs(i, bits)
{
  //right bitwise shift
  return Math.floor(i / Math.pow(2, bits));
}

function intToString(n, byteCount, bigEndian)
{
  var ret = "", bits = 0;
  for(var i = 0; i < byteCount; i++)
  {
    ret = String.fromCharCode(rs(n, bits) & 0xFF) + ret;
    bits += 8;
  }
  if(bigEndian)
  {
    var reverse = "";
    for(var i = byteCount-1; i > 0; i++)
      reverse += ret[i];
      
    ret = reverse;
  }
  return ret;
}

function parseFontHeader(data, struct)
{
  struct.add(FIXED, true); //table version number
  struct.add(FIXED, true); //fontRevision
  struct.add(ULONG, false, false, null, "checkSumAdjustment"); 
  struct.add(ULONG, true); //magic number
  struct.add(USHORT, false, false, function()
  {
    return "\0" + String.fromCharCode(rand_max(ls(1, 14) + 1));
  }); // flags
  struct.add(USHORT, false, false, function()
  {
    //unitsPerEm
    return rand_range(4, 15);
  });
  struct.add(LONGDATETIME); //created
  struct.add(LONGDATETIME); //modified
  struct.add(SHORT); //xMin
  struct.add(SHORT); //yMin
  struct.add(SHORT); //xMax
  struct.add(SHORT); //yMax
  struct.add(USHORT, false, false, function() //macStyle
  {
    return "\0" + String.fromCharCode(rand_max(ls(1, 6) + 1));
  });
  struct.add(USHORT); //lowestRecPPEM
  struct.add(SHORT, true); //fontDirectionHint
  struct.add(SHORT, false, false, function() //indexToLocFormat
  {
    return "\0" + String.fromCharCode(rand_max(2));
  });
  struct.add(SHORT, true); //glyphDataFormat
}

function parseMaxProfile(data, struct)
{
  if(getMBInt(LONG, data, struct.offset) == 0x00005000)
  {
    //version 0.5
    struct.add(FIXED, true); //table version number
    struct.add(USHORT); //numGlyphs
  }
  else
  {
    //version 1.0
    struct.add(FIXED, true); //table version number
    struct.add(USHORT); //numGlyphs
    struct.add(USHORT); //maxPoints
    struct.add(USHORT); //maxContours
    struct.add(USHORT); //maxCompositePoints
    struct.add(USHORT); //maxCompositeContours
    struct.add(USHORT, false, false, function() //maxZones
    {
      return "\0" + String.fromCharCode(rand_max(2) + 1);
    });
    struct.add(USHORT); //maxTwilightPoints
    struct.add(USHORT); //maxStorage
    struct.add(USHORT); //maxFunctionDefs
    struct.add(USHORT); //maxInstructionDefs
    struct.add(USHORT); //maxStackElements
    struct.add(USHORT); //maxSizeOfInstructions
    struct.add(USHORT); //maxComponentElements
    struct.add(USHORT); //maxComponentDepth
    
  }
}

function parseHorizHeader(data, struct)
{
  struct.add(FIXED, true); //table version number
  struct.add(FWORD); //ascender
  struct.add(FWORD); //descender
  struct.add(FWORD); //linegap
  struct.add(UFWORD); //advancewidthmax
  struct.add(FWORD); //minleftsidebearing
  struct.add(FWORD); //minrightsidebearing
  struct.add(FWORD); //xmaxextent
  struct.add(SHORT); //caretsloperise
  struct.add(SHORT); //caretsloperun
  struct.add(SHORT); //caretoffset
  struct.add(SHORT, true); //reserved
  struct.add(SHORT, true); //reserved
  struct.add(SHORT, true); //reserved
  struct.add(SHORT, true); //reserved
  struct.add(SHORT); //metricdataformat
  struct.add(USHORT); //numberofhmetrics
}

function bitCast(bits)
{
  var mult = Math.pow(2, bits.length),
    num = 0;
  for (i = 0; i < bits.length; i++)
  {
    mult = mult / 2;
    num = num + bits[i] * mult;
  }
  return num;
}


function parseOS2(data, struct)
{
  if(getMBInt(SHORT, data, struct.offset) == 0x0004)
  {
    struct.add(USHORT, true); //version
    struct.add(SHORT); //xAvgCharWidth
    struct.add(USHORT, false, false, function() //usWeightClass
    {
      var c = (rand_max(9) + 1) * 100;
      return intToString(c, SHORT);
    });
    struct.add(USHORT, false, false, function() //usWidthClass
    {
      var c = rand_max(9) + 1;
      return "\0" + String.fromCharCode(c);
    });
    struct.add(USHORT, false, false, function() //fsType
    {
      var bits = [
        0,
        rand_max(2),
        rand_max(2),
        rand_max(2),
        0,
        0,
        0,
        0,
        rand_max(2),
        rand_max(2)
      ];
      var n = bitCast(bits);
      return intToString(n, SHORT);
    });
    struct.add(SHORT); //ySubscriptXSize
    struct.add(SHORT); //ySubscriptYSize
    struct.add(SHORT); //ySubscriptXOffset
    struct.add(SHORT); //ySubscriptYOffset
    struct.add(SHORT); //ySuperscriptXSize
    struct.add(SHORT); //ySuperscriptYSize
    struct.add(SHORT); //xSuperscriptXOffset
    struct.add(SHORT); //ySuperscriptYOffset
    struct.add(SHORT); //yStrikeoutSize
    struct.add(SHORT); //yStrikeoutPosition
    struct.add(SHORT); //sFamilyClass
    struct.add(BYTE * 10); //panose[10]
    struct.add(ULONG); //ulUnicodeRange1
    struct.add(ULONG); //ulUnicodeRange2
    struct.add(ULONG); //ulUnicodeRange3
    struct.add(ULONG); //ulUnicodeRange4
    struct.add(CHAR * 4); //achVendID[4]
    struct.add(USHORT); //fsSelection
    struct.add(USHORT); //usFirstCharIndex
    struct.add(USHORT); //usLastCharIndex
    struct.add(SHORT); //sTypoAscender
    struct.add(SHORT); //sTypoDescender
    struct.add(SHORT); //sTypoLineGap
    struct.add(USHORT); //usWinAscent
    struct.add(USHORT); //usWinDescent
    struct.add(ULONG); //ulCodePageRange1
    struct.add(ULONG); //ulCodePageRange2
    struct.add(SHORT); //sxHeight
    struct.add(SHORT); //sCapHeight
    struct.add(USHORT); //usDefaultChar
    struct.add(USHORT); //usBreakChar
    struct.add(USHORT); //usMaxContext
  }
}


function makeFontStructure(data)
{
  //this function assumes the input is a valid font file :P
  var offset = 0, numTables;
  var struct = structure.create(data.length);
  struct.tableHeaders = [];
  //set collapse fn to fix checksums?
  
  fontData = data;
  
  //font header. todo: support true and typ1 for sfnt version
  if(data.substr(0, 4) == "OTTO" || data.substr == "\0\1\0\0")
  {
    numTables = ls(data.charCodeAt(4), 8) + data.charCodeAt(5); //USHORT
    offset += struct.add(FIXED, true) +
      struct.add(USHORT) + //numTables
      struct.add(USHORT) + //searchRange
      struct.add(USHORT) + //entrySelector
      struct.add(USHORT);  //rangeShift
  }
  else
    return null;
  
  for(var i = 0; i < numTables; i++)
  {
    struct.tableHeaders.push(tableHeader.create(data, struct));
  }
  
  struct.tableHeaders.sort(function(a, b)
  {
    return a.offset - b.offset;
  });
    
  var header;
  for(var i = 0; header = struct.tableHeaders[i]; i++)
  {
    
    if(header.offset - struct.offset)
      //add data that wasn't parsed
      offset += struct.add(header.offset - struct.offset, true);
    
    var start = struct.offset;
    
    
    switch(header.tag)
    {
      case "cmap":
        break;
      case "head":
        //font header
        struct.currTable = header.tag;
        parseFontHeader(data, struct);
        break;
      case "hhea":
        //horizontal header
        struct.currTable = header.tag;
        parseHorizHeader(data, struct);
        break;
      case "hmtx":
        
        break;
      case "maxp":
        //maximum profile
        struct.currTable = header.tag;
        parseMaxProfile(data, struct);
        break;
      case "name":
        break;
      case "OS/2":
        struct.currTable = header.tag;
        parseOS2(data, struct);
        break;
      case "cvt ":
        struct.currTable = header.tag;
        while(struct.offset < header.offset + header.length - FWORD + 1)
          struct.add(FWORD, false, true); //List of n values referenceable by instructions. n is the number of FWORD items that fit in the size of the table.
        break;
      case "fpgm":
        while(struct.offset < header.offset + header.length - BYTE + 1)
          struct.add(BYTE, false, true); //Instructions. n is the number of BYTE items that fit in the size of the table.
        break;
      case "glyf":
        break;
      case "CFF ":
        parseCFF(data, struct);
        
        break;
      default:
        struct.add(header.length, true);
        break;
    }
   
    
    assert(struct.offset - start < header.length && struct.currTable);
    
    //font tables are padded to be on the 4-byte boundary. according to the spec, "The length of all tables should be recorded in the table record with their actual length (not their padded length)," so we can't just rely on the header
    var adjLen = (header.length + 3) & ~3;
    if(adjLen - header.length)
      struct.add(adjLen - header.length, true);
    
    struct.currTable = null;
    
  }
  return struct;
}